import 'dart:io';
import 'package:device_info_plus/device_info_plus.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:ios_utsname_ext/extension.dart';
import 'package:dart_ipify/dart_ipify.dart';

class NUIDeviceInfo extends NUIEnt<NUIDeviceInfo>{
  NUIDeviceInfo.empty();
  NUIDeviceInfo({this.uuid, this.model, this.version, this.brand, this.manufacturer, this.ipAddress, this.details, this.platform});
  String? model;
  String? version;
  String? brand;
  String? manufacturer;
  String? platform;
  String? uuid;
  String? ipAddress;
  Map<String, dynamic>? details;

  @override
  NUIEntMapper<NUIDeviceInfo> mapper() => NUIDeviceInfoEntity();
}

class NUIDeviceInfoEntity extends NUIEntMapper<NUIDeviceInfo>{
  @override
  String entityName() => "NUIDeviceInfo";

  @override
  List<NUIEntField> fields() {
    return [
      NUIEntField<NUIDeviceInfo>(name: "model", type: NUIEntType.STRING, setter: (data, value) => data.model = value, getter: (data) => data.model),
      NUIEntField<NUIDeviceInfo>(name: "version", type: NUIEntType.STRING, setter: (data, value) => data.version = value, getter: (data) => data.version),
      NUIEntField<NUIDeviceInfo>(name: "brand", type: NUIEntType.STRING, setter: (data, value) => data.brand = value, getter: (data) => data.brand),
      NUIEntField<NUIDeviceInfo>(name: "manufacturer", type: NUIEntType.STRING, setter: (data, value) => data.manufacturer = value, getter: (data) => data.manufacturer),
      NUIEntField<NUIDeviceInfo>(name: "platform", type: NUIEntType.STRING, setter: (data, value) => data.platform = value, getter: (data) => data.platform),
      NUIEntField<NUIDeviceInfo>(name: "uuid", type: NUIEntType.STRING, setter: (data, value) => data.uuid = value, getter: (data) => data.uuid),
      NUIEntField<NUIDeviceInfo>(name: "details", type: NUIEntType.MAP, setter: (data, value) => data.details = asMap(value), getter: (data) => data.details),
    ];
  }

  @override
  NUIDeviceInfo newInstance() => NUIDeviceInfo.empty();
}

enum NUIPlatform{
  ANDROID,
  IOS,
  WEB
}

extension NUIPlatformExt on NUIPlatform{
  String value(){
    switch(this){
      case NUIPlatform.ANDROID: {
        return "Android";
      }
      case NUIPlatform.IOS: {
        return "iOS";
      }
      case NUIPlatform.WEB: {
        return "Web";
      }
      default: {
        return "Unknown";
      }
    }
  }
}

Future<NUIDeviceInfo?> getDeviceInfo() async {
  return NUIDeviceUtil.getDeviceInfo();
}

class NUIDeviceUtil{
  static final DeviceInfoPlugin deviceInfoPlugin = DeviceInfoPlugin();
  NUIDeviceUtil._builder();

  static Future<NUIDeviceInfo?> getDeviceInfo() async {
    Map<String, dynamic> deviceData;
    NUIDeviceInfo? deviceInfo;
    final ip = await Ipify.ipv4();

    try {
      if(kIsWeb){
        final webBrowserInfo = await deviceInfoPlugin.webBrowserInfo;
        deviceData = _readWebBrowserInfo(webBrowserInfo);
        deviceInfo = NUIDeviceInfo(
            model: webBrowserInfo.appCodeName,
            version: webBrowserInfo.appVersion,
            brand: webBrowserInfo.appName,
            manufacturer: webBrowserInfo.browserName.name,
            details: deviceData,
            platform: NUIPlatform.WEB.value(),
            ipAddress: ip
        );
      }else{
        if (Platform.isAndroid) {
          final androidDeviceInfo = await deviceInfoPlugin.androidInfo;
          final uuid = androidDeviceInfo.androidId;
          deviceData = _readAndroidBuildData(androidDeviceInfo);
          deviceInfo = NUIDeviceInfo(
              uuid: uuid,
              model: androidDeviceInfo.model,
              version: androidDeviceInfo.version.sdkInt.toString(),
              brand: androidDeviceInfo.brand,
              manufacturer: androidDeviceInfo.manufacturer,
              details: deviceData,
              platform: NUIPlatform.ANDROID.value(),
              ipAddress: ip
          );
        } else if (Platform.isIOS) {
          final iosDeviceInfo = await deviceInfoPlugin.iosInfo;
          final moreDetail = iosDeviceInfo.utsname;
          final machine = moreDetail.machine;
          final maskedModelName = machine?.iOSProductName ?? machine ?? iosDeviceInfo.model;
          final uuid = iosDeviceInfo.identifierForVendor;
          deviceData = _readIosDeviceInfo(iosDeviceInfo);
          deviceInfo = NUIDeviceInfo(
              uuid: uuid,
              model: maskedModelName,
              version: iosDeviceInfo.systemVersion,
              brand: "iOS",
              manufacturer: "Apple",
              details: deviceData,
              platform: NUIPlatform.IOS.value(),
              ipAddress: ip
          );
        }
      }
    } on PlatformException {
      deviceData = <String, dynamic>{
        'Error:': 'Failed to get platform version.'
      };
      deviceInfo = null;
    }
    return deviceInfo;
  }

  static Map<String, dynamic> _readAndroidBuildData(AndroidDeviceInfo build) {
    return <String, dynamic>{
      'version.securityPatch': build.version.securityPatch,
      'version.sdkInt': build.version.sdkInt,
      'version.release': build.version.release,
      'version.previewSdkInt': build.version.previewSdkInt,
      'version.incremental': build.version.incremental,
      'version.codename': build.version.codename,
      'version.baseOS': build.version.baseOS,
      'board': build.board,
      'bootloader': build.bootloader,
      'brand': build.brand,
      'device': build.device,
      'display': build.display,
      'fingerprint': build.fingerprint,
      'hardware': build.hardware,
      'host': build.host,
      'id': build.id,
      'manufacturer': build.manufacturer,
      'model': build.model,
      'product': build.product,
      'supported32BitAbis': build.supported32BitAbis,
      'supported64BitAbis': build.supported64BitAbis,
      'supportedAbis': build.supportedAbis,
      'tags': build.tags,
      'type': build.type,
      'isPhysicalDevice': build.isPhysicalDevice,
      'androidId': build.androidId,
      'systemFeatures': build.systemFeatures,
    };
  }

  static Map<String, dynamic> _readIosDeviceInfo(IosDeviceInfo data) {
    return <String, dynamic>{
      'name': data.name,
      'systemName': data.systemName,
      'systemVersion': data.systemVersion,
      'model': data.model,
      'localizedModel': data.localizedModel,
      'identifierForVendor': data.identifierForVendor,
      'isPhysicalDevice': data.isPhysicalDevice,
      'utsname.sysname:': data.utsname.sysname,
      'utsname.nodename:': data.utsname.nodename,
      'utsname.release:': data.utsname.release,
      'utsname.version:': data.utsname.version,
      'utsname.machine:': data.utsname.machine,
    };
  }

  static Map<String, dynamic> _readWebBrowserInfo(WebBrowserInfo data) {
    return <String, dynamic>{
      'browserName': describeEnum(data.browserName),
      'appCodeName': data.appCodeName,
      'appName': data.appName,
      'appVersion': data.appVersion,
      'deviceMemory': data.deviceMemory,
      'language': data.language,
      'languages': data.languages,
      'platform': data.platform,
      'product': data.product,
      'productSub': data.productSub,
      'userAgent': data.userAgent,
      'vendor': data.vendor,
      'vendorSub': data.vendorSub,
      'hardwareConcurrency': data.hardwareConcurrency,
      'maxTouchPoints': data.maxTouchPoints,
    };
  }
}