export 'nui_cryptography_exception.dart';

import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:cryptography/cryptography.dart';
import 'package:cryptography_flutter/cryptography_flutter.dart';
import 'package:nui_core/cryptograhy/nui_cryptography_exception.dart';
import 'package:nui_core/cryptograhy/nui_cryptopgrahy_bloc.dart';
import '../code/nui_code_util.dart';


abstract class NUICryptographyInterface{

  FutureOr<String> encrypt({required String data, String moduleKey});
  FutureOr<String> decrypt({required String data, String moduleKey});
  FutureOr<Map<String, dynamic>> encryptMapValue({required Map<String, dynamic> data, String moduleKey, List<String> skipForKey});
  FutureOr<Map<String, dynamic>> decryptMapValue({required Map<String, dynamic> data, String moduleKey, List<String> skipForKey});
  FutureOr<File> encryptFile({required File file, required String writeToDirectory, String moduleKey});
  FutureOr<File> decryptFile({required File file, required String writeToDirectory, String moduleKey});
  void setNewKey({required String newModule, required String newKey, bool override = true});

}

class NUICryptography implements NUICryptographyInterface{

  final String _defaultKey;
  final int _KEY_LENGTH;
  final int _KEY_ITERATIONS_COUNT;
  //Fixed variables that are same as backend
  static const int _SALT_LENGTH = 16;
  static const int _IV_LENGTH = 12;

  static final Map<String, String> module = {};
  static const String defaultModuleKey = "NUICryptoDefaultKey";

  static NUICryptography? _instance;
  late NUICryptographyBloc cryptographyBloc;

  NUICryptography._builder(NUICryptographyBuilder builder) :
        _defaultKey = builder._defaultKey ?? defaultModuleKey,
        _KEY_LENGTH = builder._KEY_LENGTH ?? 32,
        _KEY_ITERATIONS_COUNT = builder._KEY_ITERATIONS_COUNT ?? 10000{
    if(utf8.encode(_defaultKey).length != _KEY_LENGTH) throw NUICryptoException(cause: NUICryptoError.KEY_LENGTH_NOT_MATCH);
    module[defaultModuleKey] = _defaultKey;
    cryptographyBloc = NUICryptographyBloc();
  }

  static NUICryptography? get(){
    if(_instance == null) throw Exception("NUICryptography -> NUICryptography instance has not been initialized with builder");
    return _instance;
  }

  void setNewKey({required String newModule, required String newKey, bool override = true}){
    if(utf8.encode(newKey).length != _KEY_LENGTH) throw NUICryptoException(cause: NUICryptoError.KEY_LENGTH_NOT_MATCH);
    if(override){
      module[newModule] = newKey;
      return;
    }
    if(module.containsKey(newModule)) {
      throw NUICryptoException(cause: NUICryptoError.KEY_CANNOT_BE_OVERRIDE);
    }else{
      module[newModule] = newKey;
    }
  }

  FutureOr<String> _doEncryption(String data, String rawKey) async {

    int rawKeyLength = utf8.encode(rawKey).length;
    final algorithm = FlutterAesGcm(secretKeyLength: rawKeyLength, fallback: AesGcm.with256bits());
    final result = await cryptographyBloc.aes256EncryptPlain(
        algorithm: algorithm,
        rawKey: rawKey,
        rawKeyLength: rawKeyLength,
        data: data,
        iteration: _KEY_ITERATIONS_COUNT,
        saltLength: _SALT_LENGTH,
        ivLength: _IV_LENGTH
    );
    // return encrypted text
    return base64.encode(result);
  }
  
  ///Pass in base64 String as data to decrypt
  FutureOr<String> _doDecryption(String data, String rawKey) async {

    int rawKeyLength = utf8.encode(rawKey).length;
    final algorithm = FlutterAesGcm(secretKeyLength: rawKeyLength, fallback: AesGcm.with256bits());

    final cleartext = await cryptographyBloc.aes256DecryptPlain(
        algorithm: algorithm,
        rawKey: rawKey,
        rawKeyLength: rawKeyLength,
        data: data,
        iteration: _KEY_ITERATIONS_COUNT,
        saltLength: _SALT_LENGTH,
        ivLength: _IV_LENGTH
    );
    // return decrypted text
    return utf8.decode(cleartext);
  }

  FutureOr<Map<String, dynamic>> _doMapEncryption(Map<String, dynamic> mapData, String rawKey, List<String> skipForKey) async {

    Map<String, dynamic> encryptedMap = {};
    int rawKeyLength = utf8.encode(rawKey).length;
    final algorithm = FlutterAesGcm(secretKeyLength: rawKeyLength, fallback: AesGcm.with256bits());

    // Encrypted the map value
    await Future.forEach(mapData.entries, (element) async {
      // Skip key, do not need to encrypt
      if(skipForKey.contains(element.key)){
        encryptedMap[element.key] = element.value;
        return;
      }
      // Only String can be encrypted
      if(element.value is String){
        try{
          final result = await cryptographyBloc.aes256EncryptPlain(
              algorithm: algorithm,
              rawKey: rawKey,
              rawKeyLength: rawKeyLength,
              data: element.value,
              iteration: _KEY_ITERATIONS_COUNT,
              saltLength: _SALT_LENGTH,
              ivLength: _IV_LENGTH
          );
          encryptedMap[element.key] = base64.encode(result);
        }catch(e){
          //Not throwing error, just return original value
          logNUI("NUICryptography", "Encrypt failed for value [${element.value}] with error => $e");
          encryptedMap[element.key] = element.value;
        }
      }else{
        encryptedMap[element.key] = element.value;
      }
    });
    return encryptedMap;
  }

  FutureOr<Map<String, dynamic>> _doMapDecryption(Map<String, dynamic> mapData, String rawKey, List<String> skipForKey) async {

    Map<String, dynamic> decryptedMap = {};
    int rawKeyLength = utf8.encode(rawKey).length;
    final algorithm = FlutterAesGcm(secretKeyLength: rawKeyLength, fallback: AesGcm.with256bits());
    await Future.forEach(mapData.entries, (element) async {
      // Skip key, do not need to decrypt
      if(skipForKey.contains(element.key)){
        decryptedMap[element.key] = element.value;
        return;
      }
      if(element.value is String){
        try{
          final cleartext = await cryptographyBloc.aes256DecryptPlain(
              algorithm: algorithm,
              rawKey: rawKey,
              rawKeyLength: rawKeyLength,
              data: element.value,
              iteration: _KEY_ITERATIONS_COUNT,
              saltLength: _SALT_LENGTH,
              ivLength: _IV_LENGTH
          );
          decryptedMap[element.key] = utf8.decode(cleartext);
        }catch(e){
          //Not throwing error, just return original value
          logNUI("NUICryptography", "Decrypt failed for value [${element.value}] with error => $e");
          decryptedMap[element.key] = element.value;
        }
      }else{
        decryptedMap[element.key] = element.value;
      }
    });
    return decryptedMap;
  }

  FutureOr<Stream<List<int>>> _doEncryptionFile(Stream<List<int>> data, String rawKey) async {
    int rawKeyLength = utf8.encode(rawKey).length;
    final algorithm = FlutterAesGcm(secretKeyLength: rawKeyLength, fallback: AesGcm.with256bits());

    final result = await cryptographyBloc.aes256EncryptBytesAsStream(
        algorithm: algorithm,
        rawKey: rawKey,
        rawKeyLength: rawKeyLength,
        data: data
    );
    // return encrypted file
    return result;
  }

  FutureOr<Stream<List<int>>> _doDecryptionFile(Stream<List<int>> data, String rawKey) async {
    int rawKeyLength = utf8.encode(rawKey).length;
    final algorithm = FlutterAesGcm(secretKeyLength: rawKeyLength, fallback: AesGcm.with256bits());

    final result = await cryptographyBloc.aes256EncryptBytesAsStream(
        algorithm: algorithm,
        rawKey: rawKey,
        rawKeyLength: rawKeyLength,
        data: data
    );
    // return decrypted text
    return result;
  }

  /// The default value moduleKey -> 'defaultKey'
  /// 'defaultKey' is set upon initialize the NUICryptography
  @override
  FutureOr<String> encrypt({required String data, String moduleKey = defaultModuleKey}) async {
    //Info to pass into isolate
    if(!module.containsKey(moduleKey)){
      throw NUICryptoException(cause: NUICryptoError.KEY_NOT_FOUND);
    }
    
    try{
      return await _doEncryption(data, module[moduleKey]!);
    }on Exception catch(e){
      logNUI("NUICryptography in Flutter Isolate", "Encryption Exception ${e.toString()}");
      rethrow;
    }
  }

  /// The default value moduleKey -> 'defaultKey'
  /// 'defaultKey' is set upon initialize the NUICryptography
  @override
  FutureOr<String> decrypt({required String data, String moduleKey = defaultModuleKey}) async {
    //Info to pass into isolate
    if(!module.containsKey(moduleKey)){
      throw NUICryptoException(cause: NUICryptoError.KEY_NOT_FOUND);
    }

    try{
      return await _doDecryption(data, module[moduleKey]!);
    }on Exception catch(e){
      logNUI("NUICryptography in Flutter Isolate", "Encryption Exception ${e.toString()}");
      rethrow;
    }
  }

  @override
  FutureOr<Map<String, dynamic>> encryptMapValue({required Map<String, dynamic> data, String moduleKey = defaultModuleKey, List<String> skipForKey = const []}) async {
    //Info to pass into isolate
    if(!module.containsKey(moduleKey)){
      throw NUICryptoException(cause: NUICryptoError.KEY_NOT_FOUND);
    }

    try{
      return await _doMapEncryption(data, module[moduleKey]!, skipForKey);
    }on Exception catch(e){
      logNUI("NUICryptography in Flutter Isolate", "Encryption Exception ${e.toString()}");
      rethrow;
    }
  }

  @override
  FutureOr<Map<String, dynamic>> decryptMapValue({required Map<String, dynamic> data, String moduleKey = defaultModuleKey, List<String> skipForKey = const []}) async {
    //Info to pass into isolate
    if(!module.containsKey(moduleKey)){
      throw NUICryptoException(cause: NUICryptoError.KEY_NOT_FOUND);
    }

    try{
      return await _doMapDecryption(data, module[moduleKey]!, skipForKey);
    }on Exception catch(e){
      logNUI("NUICryptography in Flutter Isolate", "Encryption Exception ${e.toString()}");
      rethrow;
    }
  }

  @override
  FutureOr<File> encryptFile({required File file, required String writeToDirectory, String moduleKey  = defaultModuleKey}) async {

    bool isFileExist = await file.exists();
    if(!isFileExist){throw NUICryptoException(cause: NUICryptoError.FILE_TO_ENCRYPT_NOT_FOUND);}

    String baseFilename = file.path.split('/').last;
    const String encryptedExtension = ".encrypted";

    String writeTo = writeToDirectory + "${writeToDirectory.endsWith('/') ? "" : "/"}" + baseFilename + encryptedExtension;

    //Info to pass into isolate
    if(!module.containsKey(moduleKey)){
      throw NUICryptoException(cause: NUICryptoError.KEY_NOT_FOUND);
    }

    final sink = File(writeTo).openWrite();

    try{
      final streamData = file.openRead();
      final encryptedStream = await _doEncryptionFile(streamData, module[moduleKey]!);
      await sink.addStream(encryptedStream);

    }on Exception catch(e){
      logNUI("NUICryptography", "Encryption Exception ${e.toString()}");
      rethrow;
    }finally {
      sink.close();
      return File(writeTo);
    }


  }

  @override
  FutureOr<File> decryptFile({required File file, required String writeToDirectory, String moduleKey = defaultModuleKey}) async {
    bool isFileExist = await file.exists();
    if(!isFileExist){throw NUICryptoException(cause: NUICryptoError.FILE_TO_ENCRYPT_NOT_FOUND);}

    // file/path/to/image.jpg.encrypted -> image.jpg.encrypted
    String baseFilename = file.path.split('/').last;
    // image.jpg.encrypted -> "encrypted"
    String extension = baseFilename.split('.').last;

    // image.jpg.encrypted -> ["image","jpg","encrypted"]
    List<String> extensions = baseFilename.split('.');
    // [image","jpg","encrypted"] -> ["image","jpg"]
    extensions.removeLast();
    // ["image","jpg"] -> image.jpg
    String filename = extensions.join(".");

    if(extension != 'encrypted'){throw NUICryptoException(cause: NUICryptoError.FILE_TO_ENCRYPT_NOT_FOUND);}
    String writeTo = writeToDirectory + "${writeToDirectory.endsWith('/') ? "" : "/"}" + filename;

    //Info to pass into isolate
    if(!module.containsKey(moduleKey)){
      throw NUICryptoException(cause: NUICryptoError.KEY_NOT_FOUND);
    }

    final sink = File(writeTo).openWrite();

    try{
      final streamData = file.openRead();
      final decryptedStream = await _doDecryptionFile(streamData, module[moduleKey]!);
      await sink.addStream(decryptedStream);

    }on Exception catch(e){
      logNUI("NUICryptography in Flutter Isolate", "Encryption Exception ${e.toString()}");
      rethrow;
    }finally {
      sink.close();
      return File(writeTo);
    }
  }

}

class NUICryptographyBuilder{
  int? _KEY_LENGTH;
  int? _KEY_ITERATIONS_COUNT;
  String? _defaultKey;

  NUICryptographyBuilder keyLength(int length){
    _KEY_LENGTH = length;
    return this;
  }

  NUICryptographyBuilder iterations(int count){
    _KEY_ITERATIONS_COUNT = count;
    return this;
  }

  NUICryptographyBuilder setDefaultKey(String key){
    _defaultKey = key;
    return this;
  }

  Future<NUICryptography> build() async{
    final crypto = NUICryptography._builder(this);
    // logNUI("NUICryptography", "Initialization successful");
    NUICryptography._instance = crypto;
    return crypto;
  }

}